#include "p18F46K22.inc"
#include "config.inc"

;       
;    http://chipmk.ru/index.php/24-uchimsya-programmirovat/95-zapis-v-pamyat-programm-pic18xxxx
    

  ;errorlevel -302		; supress the 'not in bank0' warning
  ;cblock 0x00			; Access RAM

;Constants
;
  
WBUF		    equ	0x0300	;  512 byte.
INFO		    equ	0x0400	; 1048 byte.
STACK		    equ	0x0800	; 2048 byte.
	
;----------------------------------------------------------------------------
;Bit Definitions
	
SPBRG_VAL	EQU	.25		;set baud rate 9600 for 4Mhz clock
;SPBRG_VAL	EQU	.103   
	
UA_BLEN		EQU	.80 		;length of transmit circular buffer

;----------------------------------------------------------------------------
;Bit Definitions

TxBufFull	EQU	0		;bit indicates Tx buffer is full
TxBufEmpty	EQU	1		;bit indicates Tx buffer is empty
RxBufFull	EQU	2		;bit indicates Rx buffer is full
RxBufEmpty	EQU	3		;bit indicates Rx buffer is empty
ReceivedCR	EQU	4		;bit indicates <CR> character received

	
;----------------------------------------------------------------------------
	
	;Variables
	
	    CBLOCK 0x00C

	    W_TEMP			; saving space for the current context
	    W_SHADOW   
	    STATUS_TEMP	
	    STATUS_SHADOW
	    
	    FSR0L_SHADOW
	    FSR0H_SHADOW

	    Flags			;byte to store indicator flags
	    RxByte			;number of bytes received
	    TxByte			;number of bytes received

            tIntcnt    : 1		; timer delay counter (interruption)
            W_tmp      : 1		; saving space for the current context
            STATUS_tmp : 1
	
        ENDC
	
		
        ORG     0x0000
        goto    START

;----------------------------------------------------------------------------
;High priority interrupt routine.
	
	ORG	0x0008
	goto    HIGH_ISR

    	ORG     0x0018		; Low Priority Interrupt Vector
	goto    LOW_ISR
	
HIGH_ISR:

	movff	FSR0H,FSR0H_SHADOW	;save FSR0H register
	movff	FSR0L,FSR0L_SHADOW	;save FSR0L register
    
	MOVWF	W_TEMP		; Copy W to a Temporary Register
	SWAPF	STATUS,W	; Swap STATUS nibbles and place
	MOVWF	STATUS_TEMP	; Save STATUS to a Temporary register in Bank0

    	movlw   b'00000010'
        xorwf   LATE, f         ;  w = w, Reg = Reg ^ w
	
	btfsc   INTCON,T0IF	; source of interruption != Timer0
        goto    Timer0_int
	
	btfss	PIR1,RC1IF		;test for RCIF = 4 receive interrupt flag
	goto	cont2			;if RCIF is not set, done with test
	btfsc	PIE1,RC1IE		;test for RCIF = 4 receive interrupt flag
	bra 	RX_BUFFER 	

cont2:
	
HIGH_END:
    
	movff	FSR0L_SHADOW,FSR0L	;restore FSR0L register
	movff	FSR0H_SHADOW,FSR0H	;restore FSR0H register
    
	SWAPF	STATUS_TEMP,W		; Swap original STATUS register value
 	MOVWF	STATUS			; Restore STATUS register from
	SWAPF	W_TEMP,F		; Swap W_Temp nibbles and return
	SWAPF	W_TEMP,W		; Swap W_Temp to W to restore original
	RETFIE				; Return and enable interrupts

Timer0_int:
   
    	movlw   b'00000010'
        xorwf   LATE, f         ;  w = w, Reg = Reg ^ w
	
        decf    tIntcnt
        bcf     INTCON,T0IF     ; clean overflow flag

	movlw   0x00		; load start value
	movwf   TMR0H
	movwf   TMR0L		; and write the buffer
   
	SWAPF	STATUS_TEMP,W	; Swap original STATUS register value
 	MOVWF	STATUS		; Restore STATUS register from
	SWAPF	W_TEMP,F	; Swap W_Temp nibbles and return
	SWAPF	W_TEMP,W	; Swap W_Temp to W to restore original
	RETFIE			; Return and enable interrupts

RX_BUFFER:	

	btfsc	RCSTA,OERR	    ;if overrun error
	bra	ErrOERR		    ;then go handle error
	btfsc	RCSTA,FERR	    ;if framing error
	bra	ErrFERR		    ;then go handle error
	
	banksel WBUF 
	
	movf    UABufIndex,w 
	xorlw	UA_BLEN		    ;UABuffer is full?
	bnz	RX_READY
	    
	bsf	PIE1,TX1IE	    ;
	bra	HIGH_END
	
RX_READY:    
    
	incf	UABufIndex,f
	movff	UABufHiAddress,FSR0H	;Load Pointer to last byte
	movff	UABufLoAddress,FSR0L	
	
	movf	RCREG,W			;Save byte to buffer      
	movwf	POSTINC0
	
	movff	FSR0H, UABufHiAddress	;Save pointer into memory
	movff	FSR0L, UABufLoAddress	

	xorlw	0x0d		    ;compare with <CR>		
	;btfsc	STATUS,Z	    ;check if the same
	;bsf	Flags,ReceivedCR    ;indicate <CR> character received
	;xorlw	0x0d		    ;change back to valid data

	bsf	PIE1,TX1IE		; Enable transfer byte UART
	bra	HIGH_END

; can do special error handling here - this code simply clears and continues

ErrOERR:	bcf	RCSTA,CREN	; reset the receiver logic
		bsf	RCSTA,CREN	; enable reception again
		bra	HIGH_END

;error because FERR framing error bit is set
;can do special error handling here - this code simply clears and continues

ErrFERR:	movf	RCREG,W		;discard received data that has error
		bra	HIGH_END
	
;error because receive buffer is full
;can do special error handling here - this code checks and discards the data

ErrRxOver:	movf	RCREG,W		;discard received data
		xorlw	0x0d		;but compare with <CR>		
		btfsc	STATUS,Z	;check if the same
		bsf	Flags,ReceivedCR ;indicate <CR> character received
		bra	HIGH_END

;----------------------------------------------
;Add a byte (in WREG) to the end of the receive buffer.

PutRxBuffer:	
		return

		
;----------------------------------------------------------------------------
;Low priority interrupt routine.

LOW_ISR:    MOVWF   W_SHADOW		; Copy W to a Temporary Register
	    SWAPF   STATUS,W		; Swap STATUS nibbles and place
	    MOVWF   STATUS_SHADOW	; Save STATUS in Bank0

	    movlw	b'00000001'
	    xorwf	LATE, f         ;  w = w, Reg = Reg ^ w
	    
	    btfss	PIR1,TX1IF	;test for TXIF transmit interrupt flag
	    bra		LOW_END		;if TXIF is not set, done with test
	    btfsc	PIE1,TX1IE	;else test if Tx interrupt is enabled
	    bra		TX_BUFFER	;if so, go transmit data	
LOW_END:	
	    SWAPF	STATUS_SHADOW,W	; Swap original STATUS register value
	    MOVWF	STATUS		; Restore STATUS register from
	    SWAPF	W_SHADOW,F	; Swap W_Temp nibbles and return
	    SWAPF	W_SHADOW,W	; Swap W_Temp to W to restore original
	    RETFIE			; Return and enable interrupts

TX_BUFFER:
            banksel     WBUF 
	    movf        UABufIndex,w	  
	    xorlw 	 .0              ; buffer is empty?
	    bnz  	TX_BYTE

	    bcf		PIE1,TX1IE	; buffer is empty - disable interrupt
	    bra		LOW_END
TX_BYTE:  
    	    decf        UABufIndex,f	 

	    movff	UABufHiAddress,FSR0H	;put EndPointer into FSR0
	    movff	UABufLoAddress,FSR0L	;put EndPointer into FSR0

	    movf        POSTDEC0,F,ACCESS
	    movff	INDF0,TXREG

	    movff	FSR0H, UABufHiAddress	;put EndPointer into FSR0
	    movff	FSR0L, UABufLoAddress	;put EndPointer into FSR0

	    bra		LOW_END

;******************************************************************************
;       ORG     0x0008      ; High Priority Interrupt Vector
;       ORG     0x0018      ; Low Priority Interrupt Vector
;******************************************************************************	

START
   
    call    InitLEDs
    movlw   b'00000000'
    movwf   PORTE

    ;call    InitTimer0

    call  InitUABuffer
    
    rcall InitSerial
    
    GOTO $                          ; loop forever

     ;
; Configuration of the DELs

InitLEDs:
    
	clrf  TRISE 
	clrf  ANSELE
        clrf  LATE           ; turn off LEDs
        return
	
InitTimer0:
    
	;bsf     T0CON, TMR0ON
       
	; timer 0 config:   
	
        bsf     T0CON,TMR0ON	; enable Timer0
        bsf     T0CON,T08BIT	; 8 bits timer (0=16 bits; 1=8 bits)
        bcf     T0CON,T0CS	; timer source: internal timer of the processor
       ;bcf	T0CON,T0SE 
        bcf     T0CON,PSA	; select prescaler
        bcf     T0CON,T0PS2	; set prescaler scale to 1:128 (110)
        bcf     T0CON,T0PS1
        bcf     T0CON,T0PS0

	movlw   0x00		; load start value
	movwf   TMR0H
	movwf   TMR0L		; and write the buffer

	bsf	INTCON,GIEL
	bsf	INTCON,GIEH
	
        bcf     INTCON,T0IF    ; clean the overflow flag
        bsf     INTCON,TMR0IE  ; enable Timer0 interrupts
	
        return    
   
InitSerial:	

	banksel ANSELC
	bcf  	ANSELC, ANSC6	; 6 bit TX as Digital 
	bcf  	ANSELC, ANSC7	; 7 bit RX as Digital 
	
	banksel 0
	bcf  	TRISC,TX1 	; 6 bit Digital input disabled 
	bsf  	TRISC,RX1 	; 7 bit Digital input enabled  

	movlw	SPBRG_VAL	; set baud rate
	movwf	SPBRG		; SPBRG = ((int)(FOSC/(16UL * BAUD) -1));
	
	bcf	TXSTA,SYNC	; EUSART Mode 1 = Synchronous 0 = Asynchronous 
	bsf	TXSTA,TXEN	; 1 = Transmit enabled 0 = Transmit disabled 
	bcf	TXSTA,BRGH	; 1 = High speed baud rate 0 = Low speed
	bsf	TXSTA,TRMT	; Transmit Shift Register Status bit 
				; 1 = TSR empty 0 = TSR full 
	bcf	TXSTA,TX9 	; 0 - Selects 8-bit transmission

	bsf	RCSTA,SPEN	; Serial Port Enable bit
	bsf	RCSTA,SREN	; Single Receive. Asynchronous mode: Don?t care
	bsf	RCSTA,CREN	; Enable/Disable receiver 

	bcf	RCSTA,RX9D 	; 0 - Selects 8 bit of received data
	
	bsf	IPR1,RC1IP 	; Receive  Interrupt Priority bit 
	bcf	IPR1,TX1IP	; Transmit Interrupt Priority bit 
	
	bsf	PIE1,TX1IE	; Enables the transmit interrupt 
	bsf	PIE1,RC1IE 	; Receive enterrupt Enable bit

	bsf	RCON,IPEN	; Enable interrupt priorities
	bsf	INTCON,GIEL	; Enables all unmasked peripheral interrupts 
	bsf	INTCON,GIEH     ; Global Interrupt Enable bit

	MOVLW A'*'
S1	BTFSS PIR1, TXIF ; wait until the last bit is gone
	BRA S1
	MOVWF TXREG
	
	return

InitUABuffer:
    
	LFSR	0,INFO	    ; Main stack
	LFSR	1,STACK	    ; Stack of returns
    
	banksel	WBUF
	
	clrf    UABufFlag
	clrf    UABufIndex
	clrf    UABufHiAddress
	clrf    UABufLoAddress
	
	movlw	HIGH UABuffer	    ;take high address of receive buffer
	movwf	UABufHiAddress	    ;and place in receive start pointer	
	
	movlw	LOW UABuffer	;take low address of receive buffer
	movwf	UABufLoAddress	    ;and place in receive start pointer	
	
	return
	
	CBLOCK WBUF
	
	    UABufFlag
	    UABufIndex
	    UABufHiAddress
	    UABufLoAddress
	    UABuffer:UA_BLEN+1

	ENDC
	
	CBLOCK INFO
	    heap_info
	    loAdress_info
	    hiAdress_info
	ENDC
	
	CBLOCK STACK
	    heap_stack
	    loAdress_stack
	    hiAdress_stack
	ENDC
	
    END
    
    
